use crate::commands::backup::split_pointer_name;
use crate::descriptor::{Descriptor, DestPileDescriptor};
use anyhow::Context;
use chrono::{Date, DateTime, Utc};
use comfy_table::presets::UTF8_FULL;
use comfy_table::{Attribute, Cell, Color, ContentArrangement, Table};
use humansize::FileSize;
use itertools::Itertools;
use log::info;
use std::collections::{BTreeMap, BTreeSet};
use std::ffi::CString;
use std::io::Read;
use std::mem;
use std::mem::size_of;
use std::os::unix::ffi::OsStrExt;
use std::path::Path;
use yama::chunking::RecursiveUnchunker;
use yama::commands::{load_pile_descriptor, open_pile, retrieve_tree_node};
use yama::definitions::{ChunkId, RecursiveChunkRef, TreeNode};
use yama::pile::{DebugStatistics, Pile, RawPile};

// This module generates reports for a Datman system.
// Referenced Chunk IDs are counted and used to give an indication of size.
// Chunk IDs are summarised into u32s to reduce memory usage. Since the report is approximate,
// it doesn't matter if there are a few collisions (although they are still fairly unlikely to
// affect much).

#[derive(Clone)]
pub struct Report {
    pub last_source_backups: BTreeMap<String, Option<DateTime<Utc>>>,

    pub chunk_usage: BTreeMap<String, Sizes>,

    pub debug_stats: Option<DebugStatistics>,
}

#[derive(Clone, Default)]
pub struct Sizes {
    /// Total number of chunks that we refer to.
    pub total: u32,

    /// Each referred chunk is counted once here, but divided by the number of sharers.
    /// We are 'morally responsible' for this many chunks.
    pub moral: u32,

    /// Number of chunks that only we point to.
    pub unique: u32,

    /// Number of chunks for which we are the oldest (lexicographically earliest) pointer to point
    /// to those chunks.
    pub rollup: u32,
}

type CondensedChunkId = u32;

fn condense_chunk_id(chunk_id: ChunkId) -> CondensedChunkId {
    CondensedChunkId::from_be_bytes(
        chunk_id[0..size_of::<CondensedChunkId>()]
            .try_into()
            .unwrap(),
    )
}

pub fn generate_report(
    dest_pile_descriptor: &DestPileDescriptor,
    descriptor: &Descriptor,
) -> anyhow::Result<Report> {
    let pile_descriptor = load_pile_descriptor(&dest_pile_descriptor.path)?;
    let pile = open_pile(&dest_pile_descriptor.path, &pile_descriptor)?;

    let debug_stats = pile.raw_pile.debug_statistics()?;

    let mut pointers_to_parent_and_chunkids = BTreeMap::new();

    info!("Collecting chunk IDs... This will probably be slow.");
    for pointer_name in pile.list_pointers()? {
        let pointer = pile
            .read_pointer(&pointer_name)?
            .context("listed pointer doesn't exist")?;
        let root_node = retrieve_tree_node(&pile, pointer.chunk_ref)?;
        let pointer_chunk_ids = collect_chunk_ids(&pile, &root_node.node)?;
        pointers_to_parent_and_chunkids
            .insert(pointer_name, (pointer.parent_pointer, pointer_chunk_ids));
    }

    // Now we iterate in reverse order, making a list of count of Chunk IDs.
    // At the same time, we can also calculate 'rollup' sizes.
    let mut chunk_sharer_counts: BTreeMap<CondensedChunkId, u16> = BTreeMap::new();

    let mut pointer_stats: BTreeMap<String, Sizes> = BTreeMap::new();

    for pointer_name in pointers_to_parent_and_chunkids.keys().rev() {
        let deduped_chunks: BTreeSet<CondensedChunkId> =
            iter_over_all_chunkids_incl_parents(&pointers_to_parent_and_chunkids, &pointer_name)
                .collect();
        let mut rollup_count = 0;
        for chunk in deduped_chunks {
            let count = chunk_sharer_counts.entry(chunk).or_default();
            *count += 1;
            if *count == 1 {
                rollup_count += 1;
            }
        }
        let entry = pointer_stats.entry(pointer_name.to_owned()).or_default();
        entry.rollup = rollup_count;
    }

    // Now go through again and update all the stats!
    for pointer_name in pointers_to_parent_and_chunkids.keys().rev() {
        let deduped_chunks: BTreeSet<CondensedChunkId> =
            iter_over_all_chunkids_incl_parents(&pointers_to_parent_and_chunkids, &pointer_name)
                .collect();
        let mut unique_count = 0;
        let mut shared_count_by_sharers = [0u32; 256];
        let total_count = deduped_chunks.len();
        for chunk in deduped_chunks {
            let count = chunk_sharer_counts[&chunk];
            if count == 1 {
                unique_count += 1;
            } else {
                let num_sharers = (count as usize).min(256);
                shared_count_by_sharers[num_sharers - 1] += 1;
            }
        }

        let mut sharers_sum: f64 = 0.0;
        for (sharers_minus_one, count) in shared_count_by_sharers.into_iter().enumerate() {
            sharers_sum += (count as f64) / (sharers_minus_one + 1) as f64;
        }

        let entry = pointer_stats.entry(pointer_name.to_owned()).or_default();
        entry.moral = (sharers_sum.ceil() as u32) + unique_count;
        entry.unique = unique_count;
        entry.total = total_count as u32;
    }

    let mut last_backed_up = BTreeMap::new();
    for source_name in descriptor.sources.keys().cloned() {
        last_backed_up.insert(source_name, None);
    }

    for pointer_name in pointers_to_parent_and_chunkids.keys() {
        if let Some((source_name, date_time)) = split_pointer_name(&pointer_name) {
            last_backed_up.insert(source_name, Some(date_time));
        }
    }

    Ok(Report {
        last_source_backups: last_backed_up,
        chunk_usage: pointer_stats,
        debug_stats,
    })
}

// Does not filter duplicates...
fn iter_over_all_chunkids_incl_parents<'a>(
    pointers_to_parent_and_chunkids: &'a BTreeMap<
        String,
        (Option<String>, BTreeSet<CondensedChunkId>),
    >,
    pointer_name: &'a str,
) -> Box<dyn Iterator<Item = CondensedChunkId> + 'a> {
    let (parent, chunks) = &pointers_to_parent_and_chunkids[pointer_name];
    match parent {
        None => Box::new(chunks.iter().copied()),
        Some(parent) => Box::new(chunks.iter().copied().chain(
            iter_over_all_chunkids_incl_parents(pointers_to_parent_and_chunkids, &parent),
        )),
    }
}

fn collect_chunk_ids<RP: RawPile>(
    pile: &Pile<RP>,
    root: &TreeNode,
) -> anyhow::Result<BTreeSet<CondensedChunkId>> {
    let mut chunk_ids = BTreeSet::new();
    root.visit(
        &mut |tree_node, _| {
            match tree_node {
                TreeNode::NormalFile { content, .. } => {
                    collect_chunk_ids_from_chunkref(pile, content, &mut chunk_ids)?;
                }
                _ => {}
            }
            Ok(())
        },
        "".to_owned(),
    )?;
    Ok(chunk_ids)
}

fn collect_chunk_ids_from_chunkref<RP: RawPile>(
    pile: &Pile<RP>,
    chunk_ref: &RecursiveChunkRef,
    collection: &mut BTreeSet<CondensedChunkId>,
) -> anyhow::Result<()> {
    if chunk_ref.depth == 0 {
        collection.insert(condense_chunk_id(chunk_ref.chunk_id));
    } else {
        let shallower_chunk_ref = RecursiveChunkRef {
            chunk_id: chunk_ref.chunk_id,
            depth: chunk_ref.depth - 1,
        };
        let mut unchunker = RecursiveUnchunker::new(pile, shallower_chunk_ref);
        let mut next_chunk_id: ChunkId = Default::default();
        loop {
            let read = unchunker.read(&mut next_chunk_id[..])?;
            if read == 0 {
                break;
            } else if read < next_chunk_id.len() {
                unchunker.read_exact(&mut next_chunk_id[read..])?;
            }
            collection.insert(condense_chunk_id(next_chunk_id));
        }
    }
    Ok(())
}

pub fn print_report(report: &Report) -> anyhow::Result<()> {
    print_time_report(report)?;
    print_size_report(report)?;
    Ok(())
}

pub fn print_time_report(report: &Report) -> anyhow::Result<()> {
    println!("\nBackup times");
    let mut table = Table::new();
    table
        .load_preset(UTF8_FULL)
        .set_content_arrangement(ContentArrangement::DynamicFullWidth)
        .enforce_styling();

    table.set_header(vec![
        Cell::new("Source name").fg(Color::Cyan),
        Cell::new("Last backed up").fg(Color::Cyan),
    ]);

    let today = Utc::today();

    let sort_by_dates: Vec<(Option<Date<Utc>>, String)> = report
        .last_source_backups
        .iter()
        .map(|(name, datetime)| (datetime.map(|dt| dt.date()), name.to_owned()))
        .sorted()
        .collect();

    for (date, source_name) in sort_by_dates {
        match date {
            None => {
                table.add_row(vec![
                    Cell::new(source_name).fg(Color::Magenta),
                    Cell::new("NEVER").fg(Color::Red).add_attributes(vec![
                        Attribute::SlowBlink,
                        Attribute::RapidBlink,
                        Attribute::Bold,
                    ]),
                ]);
            }
            Some(date) => {
                let number_of_days = today.signed_duration_since(date).num_days();
                let num_days_human = if number_of_days > 0 {
                    format!("{number_of_days} days ago")
                } else {
                    format!("today")
                };

                let colour = if number_of_days < 2 {
                    Color::Green
                } else if number_of_days < 14 {
                    Color::Yellow
                } else {
                    Color::Red
                };

                let formatted_date = date.format("%F");

                let mut val_cell =
                    Cell::new(format!("{formatted_date} {num_days_human}")).fg(colour);
                if number_of_days > 28 {
                    val_cell = val_cell.add_attribute(Attribute::SlowBlink);
                }

                table.add_row(vec![Cell::new(source_name).fg(Color::Magenta), val_cell]);
            }
        }
    }

    println!("{table}");

    Ok(())
}

pub fn print_size_report(report: &Report) -> anyhow::Result<()> {
    println!("\nPile size");
    let mut table = Table::new();
    table
        .load_preset(UTF8_FULL)
        .set_content_arrangement(ContentArrangement::DynamicFullWidth)
        .enforce_styling();
    //.set_width(100);
    table.set_header(vec![
        Cell::new("Pointer name").fg(Color::Cyan),
        Cell::new("Rollup size").fg(Color::Magenta),
        Cell::new("Unique size").fg(Color::Magenta),
        Cell::new("Moral size").fg(Color::Magenta),
        Cell::new("Total size").fg(Color::Magenta),
    ]);

    let average_chunk_size = report
        .debug_stats
        .as_ref()
        .map(|stats| stats.total_chunk_size as f64 / stats.number_of_chunks as f64);
    for (pointer_name, sizes) in &report.chunk_usage {
        table.add_row(vec![
            Cell::new(pointer_name).fg(Color::Blue),
            Cell::new(format_size(sizes.rollup, average_chunk_size)).fg(Color::Yellow),
            Cell::new(format_size(sizes.unique, average_chunk_size)).fg(Color::Yellow),
            Cell::new(format_size(sizes.moral, average_chunk_size)).fg(Color::Yellow),
            Cell::new(format_size(sizes.total, average_chunk_size)).fg(Color::Yellow),
        ]);
    }

    println!("{table}");

    Ok(())
}

fn format_size(chunks: u32, average_chunk_size: Option<f64>) -> String {
    let est_size_suffix = average_chunk_size
        .map(|bytes_per_chunk| {
            let num_bytes = (chunks as f64 * bytes_per_chunk) as u64;
            let mut format = humansize::file_size_opts::BINARY;
            format.decimal_places = 1;
            format!(" ~{}", num_bytes.file_size(format).unwrap())
        })
        .unwrap_or_default();
    format!("{} c{}", chunks, est_size_suffix)
}

pub fn print_filesystem_space(pile_path: &Path) -> anyhow::Result<()> {
    let path_c = CString::new(pile_path.as_os_str().as_bytes()).unwrap();
    let stats = unsafe {
        let mut stats: libc::statfs = mem::zeroed();
        match libc::statfs(path_c.as_ptr(), &mut stats) {
            0 => Ok(stats),
            other => Err(std::io::Error::from_raw_os_error(other)),
        }
    }?;

    // On a BTRFS system with 2 disks in RAID1, note (about df -h):
    // - 'Size' shows the average size of the two disks. I think of it as 'ideal size'.
    // - 'Avail' seems to show the actual number of bytes usable.
    // - 'Used' seems to show the actual number of bytes used.
    // In short: probably avoid relying on 'size'.

    let block_size = stats.f_bsize as i64;
    let used_bytes = (stats.f_blocks - stats.f_bfree) as i64 * block_size;
    let avail_bytes = stats.f_bavail as i64 * block_size;
    let usable_bytes = used_bytes + avail_bytes;
    let theoretical_size = stats.f_blocks as i64 * block_size;

    let mut format = humansize::file_size_opts::BINARY;
    format.decimal_places = 1;
    format.decimal_zeroes = 1;

    println!("\nFilesystem Information");

    let mut table = Table::new();
    table
        .load_preset(UTF8_FULL)
        .set_content_arrangement(ContentArrangement::DynamicFullWidth)
        .enforce_styling();
    //.set_width(100);
    table.set_header(vec![
        Cell::new("Theoretical Size").fg(Color::Cyan),
        Cell::new("Usable Size").fg(Color::Cyan),
        Cell::new("Used").fg(Color::Cyan),
        Cell::new("Available").fg(Color::Cyan),
    ]);

    let available_space_colour = if avail_bytes < 8 * 1024 * 1024 * 1024 {
        Color::Red
    } else if avail_bytes < 64 * 1024 * 1024 * 1024 {
        Color::Yellow
    } else {
        Color::Green
    };

    table.add_row(vec![
        Cell::new(format!(
            "{:>9}",
            theoretical_size.file_size(&format).unwrap()
        ))
        .fg(Color::Blue),
        Cell::new(format!("{:>9}", usable_bytes.file_size(&format).unwrap())).fg(Color::Blue),
        Cell::new(format!("{:>9}", used_bytes.file_size(&format).unwrap())).fg(Color::Blue),
        Cell::new(format!("{:>9}", avail_bytes.file_size(&format).unwrap()))
            .fg(available_space_colour),
    ]);

    print!("{table}");

    Ok(())
}
